-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------

--------------------------------------------------------------------------------
-- DKP-type specific variables

local FLASK_MULTIPLIER = 0.4
local ELIXIR_MULTIPLIER = 0.2
local FOOD_MULTIPLIER = 0.1

local flasks = {
	["Flask of Arcane Fortification"] = true,
	["Chromatic Resistance"] = true,
	["Distilled Wisdom"] = true,
	["Flask of Fortification"] = true,
	["Flask of Mighty Restoration"] = true,
	["Flask of Relentless Assault"] = true,
	["Flask of Shadow Fortification"] = true,
	["Supreme Power"] = true,
	["Flask of Pure Death"] = true,
	["Flask of the Titans"] = true,
	["Supreme Power of Shattrath"] = true,
	["Fortification of Shattrath"] = true,
	["Mighty Restoration of Shattrath"] = true,
	["Relentless Assault of Shattrath"] = true,
	["Pure Death of Shattrath"] = true,
	["Blinding Light of Shattrath"] = true,
	["Unstable Flask of the Bandit"] = true,
	["Unstable Flask of the Beast"] = true,
	["Unstable Flask of the Elder"] = true,
	["Unstable Flask of the Sorcerer"] = true,
	["Unstable Flask of the Soldier"] = true,
	["Unstable Flask of the Physician"] = true,
}

local elixirs = {
	["Healing Power"] = true,
	["Elixir of Draenic Wisdom"] = true,
	["Elixir of Brute Force"] = true,
	["Fel Strength Elixir"] = true,
	["Empowerment"] = true,
	["Greater Mana Regeneration"] = true,
	["Major Shadow Power"] = true,
	["Major Armor"] = true,
	["Elixir of Ironskin"] = true,
	["Major Agility"] = true,
	["Earthen Elixir"] = true,
	["Major Frost Power"] = true,
	["Elixir of Mastery"] = true,
	["Elixir of Major Fortitude"] = true,
	["Major Strength"] = true,
	["Adept's Elixir"] = true,
	["Onslaught Elixir"] = true,
	["Greater Arcane Elixir"] = true,
	["Elixir of the Mongoose"] = true,
	["Greater Firepower"] = true,
	["Shadow Power"] = true,
	["Elixir of the Giants"] = true,
	["Arcane Elixir"] = true,
	["Elixir of Demonslaying"] = true,
}

local dkpCacheUpdated = {}

local lootDialogOpen = nil
local lootQueue = {}

local hostOffline = nil

local halfHour = 60 * 30
local oneHour = halfHour * 2

local counterRunning = nil

--------------------------------------------------------------------------------
--

-- Basically a copy of the list I made for NRT.
local ignoredItems = {
	[30316] = true, -- Devastation
	[30317] = true, -- Cosmic Infuser
	[30312] = true, -- Infinity Blade
	[30318] = true, -- Netherstrand Longbow
	[30313] = true, -- Staff of Disintegration
	[30311] = true, -- Warp Slicer
	[30320] = true, -- Bundle of Nether Spikes
	[30314] = true, -- Phaseshift Bulwark
	[30319] = true, -- Nether Spikes
	[29434] = true, -- Badge of Justice
	[22450] = true, -- Void Crystal
	[20725] = true, -- Nexus Crystal

	[32897] = true, -- Mark of the Illidari
	[32428] = true, -- Heart of Darkness
	[34664] = true, -- Sunmote

	[32230] = true, -- Shadowsong Amethyst
	[32249] = true, -- Seaspray Emerald
	[32231] = true, -- Pyrestone
	[32229] = true, -- Lionseye
	[32228] = true, -- Empyrean Sapphire
	[32227] = true, -- Crimson Spinel
}

local _G = getfenv(0)

local deformat = AceLibrary("Deformat-2.0")
local icon = LibStub("LibDBIcon-1.0")

local db = nil

local guildMemberList = {}
local hexColors = {
	WTF = "|cffa0a0a0"
}
for k, v in pairs(RAID_CLASS_COLORS) do
	hexColors[k] = "|cff" .. string.format("%02x%02x%02x", v.r * 255, v.g * 255, v.b * 255)
end
local coloredName = setmetatable({}, {__index =
	function(self, key)
		if type(key) == "nil" then return nil end
		local class = guildMemberList[key] or select(2, UnitClass(key)) or "WTF"
		self[key] = hexColors[class]  .. key .. "|r"
		return self[key]
	end
})

local updateDisplayedNames = true

local guildRanks = {}
local guildMemberRanks = {}
local guildMemberClasses = {}
local guildClasses = {}

local takingBids = nil
local tmpBidTable = {}
local lastDKPUpdate = nil
local bidders = {}
local playerName = UnitName("player")
local chatHandlers = {}

-- Thanks, Wobin.
local function round(n, shift)
	shift = 10 ^ shift
	return math.floor((n * shift) + 0.5) / shift
end

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = setmetatable({}, {__index =
	function(self, key)
		self[key] = key
		return key
	end
})

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

BananaDKP = LibStub("AceAddon-3.0"):NewAddon("BananaDKP", "AceTimer-3.0", "AceEvent-3.0", "AceComm-3.0", "AceSerializer-3.0")
local BananaDKP = BananaDKP

local function print(t)
	DEFAULT_CHAT_FRAME:AddMessage("BDKP: " .. t)
end

local function resetClient()
	if counterRunning then
		BananaDKP:StopCounter()
	end
	for k in pairs(dkpCacheUpdated) do
		dkpCacheUpdated[k] = nil
	end
	lastDKPUpdate = time()
end

local options = {
	type = "group",
	handler = BananaDKP,
	args = {
		autoHeader = {
			type = "header",
			name = L["Automatic"],
			order = 100,
		},
		start = {
			type = "execute",
			name = L["Start"],
			desc = L["Start the automatic hour counter."],
			func = "StartCounter",
			order = 101,
			disabled = function() return not UnitInRaid("player") or counterRunning end,
		},
		stop = {
			type = "execute",
			name = L["Stop"],
			desc = L["Stop the automatic hour counter."],
			func = "StopCounter",
			disabled = function() return not counterRunning end,
			order = 103,
		},
		spacer = {
			type = "header",
			name = " ",
			order = 150,
		},
		manualHeader = {
			type = "header",
			name = L["Manual"],
			order = 200,
		},
		plusone = {
			type = "text",
			name = L["Increment"],
			desc = L["Increment the hour counter for a player or the entire raid+sitout list (*) by one."],
			usage = L["<player or *>"],
			get = false,
			set = "IncrementHourCounter",
			order = 201,
		},
		minusone = {
			type = "text",
			name = L["Decrement"],
			desc = L["Decrement the hour counter for a player or the entire raid+sitout list (*) by one."],
			usage = L["<player or *>"],
			get = false,
			set = "DecrementHourCounter",
			order = 202,
		},
		spacer2 = {
			type = "header",
			name = " ",
			order = 250,
		},
		optionsHeader = {
			type = "header",
			name = L["Options"],
			order = 300,
		},
		role = {
			type = "group",
			name = L["Role"],
			desc = L["Sets your role, either client or server."],
			args = {
				client = {
					type = "toggle",
					name = L["Client"],
					desc = L["Sets this BananaDKP to act as a client."],
					get = function() return db.mode == 2 end,
					set = function(v)
						db.mode = v and 2 or 1
						resetClient()
					end,
				},
				server = {
					type = "toggle",
					name = L["Server"],
					desc = L["Sets this BananaDKP to act as a server."],
					get = function() return db.mode == 1 end,
					set = function(v)
						db.mode = v and 1 or 2
						resetClient()
					end,
				},
				host = {
					type = "text",
					name = L["Host"],
					desc = L["If acting as a client, set the host (player name) that is running the server here. Case sensitive."],
					get = function() return db.host end,
					set = function(v)
						db.host = v
						resetClient()
					end,
					usage = "<player>",
					disabled = function() return db.mode ~= 2 end,
				},
			},
			order = 301,
		},
		sitouts = {
			type = "group",
			name = L["Sitouts"],
			desc = L["List of players sitting out this raid who should still receive DKP. Click on a name to remove him from the list."],
			disabled = function() return type(currentLeader) ~= "boolean" or not currentLeader end,
			args = {
				add = {
					type = "text",
					name = L["Add player"],
					desc = L["Adds a player to the sitout list."],
					get = false,
					usage = "<player name>",
					set = "AddSitoutPlayer",
					order = 1,
				},
				clear = {
					type = "execute",
					name = L["Clear list"],
					desc = L["Clear the sitout list."],
					func = "ClearSitoutList",
					order = 2,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 3,
				},
			},
			order = 302,
			disabled = function()
				if not UnitInRaid("player") then return true end
				if not IsRaidLeader() and not IsRaidOfficer() then return true end
				return false
			end,
		},
		ranks = {
			type = "text",
			name = L["Ranks"],
			desc = L["Unchecked guild ranks will not show in the tooltip. People with these ranks will continue to earn and spend DKP as per usual, they will just be hidden from the display."],
			get = function(key)
				return not db.disallowRanks[key]
			end,
			set = function(key, val)
				db.disallowRanks[key] = not val
				if val then
					print(L["Members of rank %q will show in the tooltip."]:format(key))
				else
					print(L["Members of rank %q will NOT show in the tooltip."]:format(key))
				end
			end,
			order = 400,
			multiToggle = true,
			validate = guildRanks,
			disabled = function() return not IsInGuild() end,
		},
		classes = {
			type = "text",
			name = L["Classes"],
			desc = L["Unchecked classes will not be shown in the tooltip. Players of these classes will continue to earn and spend DKP as per usual, they will just be hidden from the display."],
			get = function(key)
				return not db.disallowClasses[key]
			end,
			set = function(key, val)
				db.disallowClasses[key] = not val
				if val then
					print(L["Members of class %q will show in the tooltip."]:format(key))
				else
					print(L["Members of class %q will NOT show in the tooltip."]:format(key))
				end
			end,
			order = 401,
			multiToggle = true,
			validate = guildClasses,
			disabled = function() return not IsInGuild() end,
		},
		minimap = {
			type = "toggle",
			name = L["Minimap icon"],
			desc = L["Toggle the minimap icon."],
			get = function() return not db.minimap.hide end,
			set = function(v)
				local hide = not v
				db.minimap.hide = hide
				if hide then
					icon:Hide("BananaDKP")
					print(L["BananaDKP minimap icon hidden, you can toggle it back with /bananadkp."])
				else
					icon:Show("BananaDKP")
				end
			end,
			order = 402,
			disabled = function() return not icon end,
		},
	},
}

local dew = AceLibrary("Dewdrop-2.0")
local LDB = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("BananaDKP", {
	type = "data source",
	text = "BDKP",
	icon = "Interface\\Icons\\Ability_Hunter_AspectOfTheMonkey",
	OnClick = function(self, button)
		if button == "RightButton" then
			dew:Open(self,
				"children", function()
					dew:FeedAceOptionsTable(options)
				end
			)
		elseif hostOffline then
			print(L["Will try to contact the host again now ..."])
			hostOffline = nil
			for k in pairs(dkpCacheUpdated) do
				dkpCacheUpdated[k] = nil
			end
		elseif UnitInRaid("player") and db.mode == 1 then
			if counterRunning then
				BananaDKP:StopCounter()
			else
				BananaDKP:StartCounter()
			end
		end
	end
})

-----------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------

local function chatFilter()
	return chatHandlers[event](arg1, arg2)
end

local chatFilters = {
	CHAT_MSG_WHISPER = chatFilter,
	CHAT_MSG_WHISPER_INFORM = chatFilter,
	CHAT_MSG_RAID = chatFilter,
	CHAT_MSG_RAID_LEADER = chatFilter,
	CHAT_MSG_SYSTEM = chatFilter,
}

-- player is the person who looted, input is the text in the input box
local function addItemFromDialog(input, player, itemLink)
	local looter = player
	if type(input) == "string" and input:trim():len() > 2 then
		if UnitExists(input) then
			looter = input:gsub("^%l", string.upper)
		else
			print(L["%q is not in your group."]:format(input))
			return false
		end
	end
	BananaDKP:AddItemToItemTable(looter, itemLink)
	return true
end

local function getDKPMultiplier(player)
	local multiplier = 1
	local buffIndex = 1
	local buff = UnitBuff(player, buffIndex)
	while buff do
		if flasks[buff] then
			multiplier = multiplier + FLASK_MULTIPLIER
		elseif buff == L["Well Fed"] or buff == L["\"Well Fed\""] then
			multiplier = multiplier + FOOD_MULTIPLIER
		elseif elixirs[buff] then
			multiplier = multiplier + ELIXIR_MULTIPLIER
		end
		buffIndex = buffIndex + 1
		buff = UnitBuff(player, buffIndex)
	end
	return multiplier
end

do
	local dbDefaults = {
		sitouts = {}, -- Remember this across sessions.
		mode = 2, -- 1 = server, 2 = client
		host = nil, -- No host set by default.
		disallowRanks = {},
		disallowClasses = {},
		dkp = {},
		itemsLooted = {},
		hours = {},
		totalRaidHours = 0,
		minimap = {},
	}
	function BananaDKP:OnInitialize()
		if not _G.BananaDKPDB then _G.BananaDKPDB = {} end
		db = _G.BananaDKPDB
		for k, v in pairs(dbDefaults) do
			if not db[k] then db[k] = v end
		end

		_G["SlashCmdList"]["BDKP_SHORTHAND"] = function()
			if db.minimap.hide then
				icon:Show("BananaDKP")
				db.minimap.hide = nil
			else
				icon:Hide("BananaDKP")
				db.minimap.hide = true
			end
		end
		_G["SLASH_BDKP_SHORTHAND1"] = "/bananadkp"

		icon:Register("BananaDKP", LDB, db.minimap)

		if IsInGuild() then GuildRoster() end
	end
end

function BananaDKP:OnEnable()
	local popup = _G.StaticPopupDialogs
	if type(popup) ~= "table" then
		popup = {}
	end
	--[[ Event order for the popup
		OnShow
		EscapePressed
		Hide

		Show
		Cancel
		Hide

		Show
		Accept
		Hide

		Show
		EnterPressed
		Hide
	]]
	if type(popup["BananaDKPItemQuery"]) ~= "table" then
		popup["BananaDKPItemQuery"] = {
			button1 = L["Add"],
			button2 = L["Ignore"],
			sound = "AuctionWindowClose",
			whileDead = 1,
			timeout = 0,
			OnShow = function()
				_G[this:GetName().."EditBox"]:SetText("")
			end,
			OnHide = function()
				_G[this:GetName().."EditBox"]:SetText("")
			end,
			OnCancel = function()
				self:IgnoreItem()
			end,
			EditBoxOnEscapePressed = function()
				this:GetParent():Hide()
				self:IgnoreItem()
			end,
			EditBoxOnEnterPressed = function(self, data)
				local input = _G[this:GetParent():GetName().."EditBox"]:GetText()
				if addItemFromDialog(input, unpack(data)) then
					this:GetParent():Hide()
				end
			end,
			OnAccept = function(self, data)
				local input = _G[this:GetParent():GetName().."EditBox"]:GetText()
				if not addItemFromDialog(input, unpack(data)) then
					return true
				end
			end,
			hasEditBox = 1,
		}
	end

	for event, filter in pairs(chatFilters) do
		ChatFrame_AddMessageEventFilter(event, filter)
	end

	self:RegisterEvent("GUILD_ROSTER_UPDATE")
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")

	-- The comm code needs some serious cleanup and refactoring now, I just hacked things on it to get it done.

	-- Multiple values (serialized)
	self:RegisterComm("BDKPM", "OnComm")
	self:RegisterComm("BDKPRM", "OnComm")
	-- Single values
	self:RegisterComm("BDKP", "OnComm")
	self:RegisterComm("BDKPR", "OnComm")
	
	self:RegisterComm("BDKPI", "OnComm")
	self:RegisterComm("BDKPH", "OnComm")

	self:RegisterComm("BDKP_SERVER_AVAILABLE", "OnComm")
	self:RegisterComm("BDKP_CLIENT_AVAILABLE", "OnComm")
	self:RegisterComm("BDKP_FULL_UPDATE", "OnComm")
	self:RegisterComm("BDKP_INCREMENT", "OnComm")
	self:RegisterComm("BDKP_DECREMENT", "OnComm")

	self:RebuildSitoutList()

	if db.mode == 1 and IsInGuild() then
		self:SendCommMessage("BDKP_SERVER_AVAILABLE", (UnitName("player")), "GUILD")
	end
	if db.mode == 2 and db.host then
		self:SendCommMessage("BDKP_CLIENT_AVAILABLE", (UnitName("player")), "WHISPER", db.host)
	end
end

-----------------------------------------------------------------------
-- Events
-----------------------------------------------------------------------

-----------------------------------------------------------------------
-- CHAT_MSG_ADDON

local clients = {}
local function broadcastSingle(player)
	local dkp = db.dkp[player] or 0
	if GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0 then
		BananaDKP:SendCommMessage("BDKP", player .. " " .. tostring(dkp), "RAID")
	end
	for k in pairs(clients) do
		if not inGroup(k) then
			BananaDKP:SendCommMessage("BDKP", player .. " " .. tostring(dkp), "WHISPER", k)
		end
	end
end

local function addItem(player)
	-- DKP -25%
	local dkp = db.dkp[player] or 0
	local price = 0
	if dkp > 0 then
		price = dkp * 25 / 100
	end
	local newDkp = dkp - price
	db.dkp[player] = newDkp

	broadcastSingle(player)

	local l = db.itemsLooted[player] or 0
	db.itemsLooted[player] = l + 1

	lastDKPUpdate = time()
	return price
end

local function incrementByOne(input)
	if db.dkp[input] then
		db.dkp[input] = db.dkp[input] + 1
	else
		db.dkp[input] = 1
	end
	if db.hours[input] then
		db.hours[input] = db.hours[input] + 1
	else
		db.hours[input] = 1
	end
end

local function decrementByOne(input)
	if db.dkp[input] then
		db.dkp[input] = db.dkp[input] - 1
	end
end

local function inGroup(player)
	return UnitInRaid(player) or UnitInParty(player) and true or nil
end

-- Max addon message length is 255, and the max name length for WoW is 12.
-- We need 14 chars per name for name + space + separator, and DKP values
-- can be quite long, so we'll allocate 10 chars per return value.
-- So if we query for 10 players at a time, we get 100 + 140 = 240 which leaves
-- 14 chars for the prefix and any random overhead.

local dkpQueryList = {}
local function actuallyQueryDKP()
	if db.host and not hostOffline then
		local sz = BananaDKP:Serialize(dkpQueryList)
		BananaDKP:SendCommMessage("BDKPRM", sz, "WHISPER", db.host)
	end
	for k in pairs(dkpQueryList) do dkpQueryList[k] = nil end
end

local queryTimer = nil
local function queueDkpQuery(player)
	table.insert(dkpQueryList, player)
	BananaDKP:CancelTimer(queryTimer, true)
	if #dkpQueryList < 10 then
		queryTimer = BananaDKP:ScheduleTimer(actuallyQueryDKP, 1)
	else
		actuallyQueryDKP()
	end
end

local function getDKP(name)
	if not name then return end
	if db.mode == 2 and not dkpCacheUpdated[name] then queueDkpQuery(name) end
	if not db.dkp[name] then return end
	return round(db.dkp[name], 2)
end

local function dkpSorter(a, b)
	local aDKP = getDKP(a) or 0
	local bDKP = getDKP(b) or 0
	return aDKP > bDKP and true or false
end

local fullUpdateThisSession = nil
local updatedClients = {}

local function fullUpdate()
	local full = BananaDKP:Serialize(db.dkp)
	if GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0 then
		BananaDKP:SendCommMessage("BDKPM", full, "RAID")
	end
	for k in pairs(clients) do
		if not inGroup(k) then
			BananaDKP:SendCommMessage("BDKPM", full, "WHISPER", k)
		end
	end
end

function BananaDKP:OnComm(prefix, message, distribution, sender)
	if not sender or not prefix or not message then return end
	if prefix == "BDKP_SERVER_AVAILABLE" and db.mode == 2 and sender == db.host then
		if not fullUpdateThisSession then
			self:SendCommMessage("BDKP_FULL_UPDATE", (UnitName("player")), "WHISPER", db.host)
			fullUpdateThisSession = true
			hostOffline = nil
		end
	elseif prefix == "BDKP_CLIENT_AVAILABLE" and db.mode == 1 then
		if not guildMemberList[message] then
			clients[message] = true
		end
		if not updatedClients[message] then
			local full = self:Serialize(db.dkp)
			if inGroup(message) then
				self:SendCommMessage("BDKPM", full, "RAID")
			else
				self:SendCommMessage("BDKPM", full, "WHISPER", message)
			end
			updatedClients[message] = true
		end
	elseif prefix == "BDKP_FULL_UPDATE" and db.mode == 1 then
		local full = self:Serialize(db.dkp)
		if inGroup(sender) then
			self:SendCommMessage("BDKPM", full, "RAID")
		else
			self:SendCommMessage("BDKPM", full, "WHISPER", sender)
		end
	elseif prefix == "BDKP_INCREMENT" and db.mode == 1 then
		local good, deSz = self:Deserialize(message)
		if not good then
			print("Failure to deserialize incoming data from " .. sender .. ".")
			return
		end
		for k, v in pairs(deSz) do
			db.dkp[k] = db.dkp[k] and (db.dkp[k] + v) or v
		end
	elseif prefix == "BDKP_DECREMENT" and db.mode == 1 then
		local good, deSz = self:Deserialize(message)
		if not good then
			print("Failure to deserialize incoming data from " .. sender .. ".")
			return
		end
		for k in pairs(deSz) do
			db.dkp[k] = db.dkp[k] and (db.dkp[k] - 1) or -1
		end
	elseif prefix == "BDKPI" and db.mode == 1 then
		if inGroup(sender) then return end -- Ignore anything from our own group.
		local player, itemLink = select(3, message:find("(%S+) (.*)"))
		if not player and itemLink then return end
		local price = addItem(player)
		print(L["%s says %s looted %s, %d has been deducted from his DKP."]:format(sender, player, itemLink, price or 0))
	elseif prefix == "BDKPH" and db.mode == 1 then
		if inGroup(sender) then return end -- Ignore anything from our own group.

		local input, action = select(3, message:find("(%S+) (.*)"))
		if action == "+" then
			incrementByOne(input)
			print(L["%s manually incremented the DKP for %s by +1."]:format(sender, input))
		elseif action == "-" then
			decrementByOne(input)
			print(L["%s manually decremented the DKP for %s by -1."]:format(sender, input))
		elseif tonumber(action) then
			action = tonumber(action)
			if not db.dkp[input] then
				db.dkp[input] = action
			else
				db.dkp[input] = db.dkp[input] + action
			end
		end
		broadcastSingle(input)
		lastDKPUpdate = time()
	elseif (prefix == "BDKP" or prefix == "BDKPM") and db.mode == 2 and sender == db.host then
		if prefix == "BDKPM" then
			local good, deSz = self:Deserialize(message)
			if not good then
				print("Failure to deserialize incoming data from " .. sender .. ".")
				return
			end
			for k, v in pairs(deSz) do
				dkpCacheUpdated[k] = true
				db.dkp[k] = v
			end
		else
			local player, dkp = select(3, message:find("(%S+) (.*)"))
			dkpCacheUpdated[player] = true
			db.dkp[player] = tonumber(dkp)
		end
		lastDKPUpdate = time()
	elseif (prefix == "BDKPR" or prefix == "BDKPRM") and db.mode == 1 then
		if prefix == "BDKPRM" then
			local good, deSz = self:Deserialize(message)
			if not good then
				print("Failure to deserialize incoming data from " .. sender .. ".")
				return
			end
			local r = {}
			for i, v in ipairs(deSz) do
				local dkp = getDKP(v)
				if dkp and dkp > 0 then
					r[v] = dkp
				end
			end
			local sz = self:Serialize(r)
			wipe(r); r = nil
			if inGroup(sender) then
				self:SendCommMessage("BDKPM", sz, "RAID")
			else
				self:SendCommMessage("BDKPM", sz, "WHISPER", sender)
			end
		else
			local dkp = getDKP(message)
			if dkp and dkp > 0 then
				if inGroup(sender) then
					self:SendCommMessage("BDKP", message .. " " .. tostring(dkp), "RAID")
				else
					self:SendCommMessage("BDKP", message .. " " .. tostring(dkp), "WHISPER", sender)
				end
			end
		end
	end
end

-----------------------------------------------------------------------
-- Ticker handling

function BananaDKP:StartCounter()
	assert(not counterRunning, "Counter is already running.")

	math.randomseed(time())
	local r = math.random(5, 30)
	counterRunning = self:ScheduleTimer("IncrementHourCounter", r * 60)
	print(L["Automatic hour counting has started, the first increment will be in ~%d minutes."]:format(r))
	self:Chat(L["BDKP hour counter started, whisper me 'sitout' if you are not in the raid but available on standby."], "GUILD")
	for player in pairs(db.sitouts) do
		self:Chat(L["BDKP: You are on sitout."], "WHISPER", nil, player)
	end
	
	self:RegisterEvent("CHAT_MSG_LOOT")
end

function BananaDKP:StopCounter()
	if not counterRunning then return end
	self:CancelTimer(counterRunning)
	counterRunning = nil
	print(L["Automatic hour counting stopped."])
	
	self:UnregisterEvent("CHAT_MSG_LOOT")
end

-----------------------------------------------------------------------
-- Item handling

function BananaDKP:ItemLooted(player, itemLink)
	if not counterRunning then return end
	if db.mode == 2 and db.host and inGroup(db.host) then return end

	if lootDialogOpen then
		table.insert(lootQueue, {player, itemLink})
	else
		lootDialogOpen = true
		StaticPopupDialogs["BananaDKPItemQuery"].text = L["%s looted %s, click Add to add it to his loot table, or enter the name of someone else who should pay for it in the text box below."]:format(player, itemLink)
		local popup = StaticPopup_Show("BananaDKPItemQuery")
		if popup then
			popup.data = {player, itemLink}
		end
	end
end

local function timerLoot(tbl)
	BananaDKP:ItemLooted(unpack(tbl))
end

local function popLootQueue()
	if #lootQueue > 0 then
		local n = table.remove(lootQueue, 1)
		BananaDKP:ScheduleTimer(timerLoot, 1, n)
	end
end

function BananaDKP:AddItemToItemTable(player, itemLink)
	lootDialogOpen = nil

	-- mode 1 = server, 2 = client
	if db.mode == 1 then
		local price = addItem(player)
		print(L["%s looted %s, %d has been deducted from his DKP."]:format(player, itemLink, price or 0))
	elseif db.mode == 2 then
		if not db.host or hostOffline then return end
		self:SendCommMessage("BDKPI", player .. " " .. itemLink, "WHISPER", db.host)
		print(L["%s looted %s, it has been added to his item table."]:format(player, itemLink))
	end

	popLootQueue()
end

function BananaDKP:IgnoreItem()
	lootDialogOpen = nil
	popLootQueue()
end

-----------------------------------------------------------------------
-- Hour handling

function BananaDKP:IncrementHourCounter(input)
	if type(input) == "string" and not input:find("*") then
		input = input:trim():gsub("^%l", string.upper)
		if db.mode == 1 then
			incrementByOne(input)
			broadcastSingle(input)
		elseif db.mode == 2 then
			if not db.host or hostOffline then return end
			self:SendCommMessage("BDKPH", input .. " +", "WHISPER", db.host)
		end
		print(L["The hour counter has been incremented by 1 for %s."]:format(input))
	else
		if type(input) == "string" and input ~= "*" then
			print(L["Can't increment the hour counter for %q."]:format(input))
			return
		end
		local dkpTable = {}
		for i = 1, GetNumRaidMembers() do
			local n = GetRaidRosterInfo(i)
			dkpTable[n] = 1 * getDKPMultiplier(n)
		end
		for k, v in pairs(db.sitouts) do
			if not dkpTable[k] then dkpTable[k] = 1 end
		end
		if db.mode == 1 then
			for k, v in pairs(dkpTable) do
				if not db.dkp[k] then
					db.dkp[k] = v
				else
					db.dkp[k] = db.dkp[k] + v
				end
				if not db.hours[k] then
					db.hours[k] = 1
				else
					db.hours[k] = db.hours[k] + 1
				end
			end
			fullUpdate()
		elseif db.mode == 2 then
			if not db.host or hostOffline then return end
			local sz = self:Serialize(dkpTable)
			self:SendCommMessage("BDKP_INCREMENT", sz, "WHISPER", db.host)
		end
		print(L["The hour counter has been incremented by 1 for all players in the raid and on the sitout list."])
		if type(input) ~= "string" then
			counterRunning = self:ScheduleTimer("IncrementHourCounter", oneHour)
			db.totalRaidHours = db.totalRaidHours + 1
		end
	end

	lastDKPUpdate = time()
end

function BananaDKP:DecrementHourCounter(input)
	if type(input) == "string" and not input:find("*") then
		input = input:trim():gsub("^%l", string.upper)
		if db.mode == 1 then
			decrementByOne(input)
			broadcastSingle(input)
		elseif db.mode == 2 then
			if not db.host or hostOffline then return end
			self:SendCommMessage("BDKPH", input .. " -", "WHISPER", db.host)
		end
		print(L["The hour counter has been decremented by 1 for %s."]:format(input))
	else
		if type(input) == "string" and input ~= "*" then
			print(L["Can't decrement the hour counter for %q."]:format(input))
			return
		end
		local dkpTable = {}
		for i = 1, GetNumRaidMembers() do
			local n = GetRaidRosterInfo(i)
			dkpTable[n] = true
		end
		for k, v in pairs(db.sitouts) do
			dbkTable[k] = true
		end
		if db.mode == 1 then
			for k, v in pairs(dkpTable) do
				if db.dkp[k] then
					db.dkp[k] = db.dkp[k] - 1
				else
					db.dkp[k] = -1
				end
			end
			fullUpdate()
		elseif db.mode == 2 then
			if not db.host or hostOffline then return end
			local sz = self:Serialize(dkpTable)
			self:SendCommMessage("BDKP_DECREMENT", sz, "WHISPER", db.host)
		end
		print(L["The hour counter has been decremented by 1 for all players in the raid and on the sitout list."])
	end

	lastDKPUpdate = time()
end

--------------------------------------------------------------------------------

function BananaDKP:CHAT_MSG_LOOT(msg)
	local player, item = deformat(msg, LOOT_ITEM)
	if not player then
		item = deformat(msg, LOOT_ITEM_SELF)
		if item then
			player = playerName
		end
	end

	if type(item) == "string" and type(player) == "string" then
		local itemName, itemLink, itemRarity = GetItemInfo(item)
		if not itemRarity or itemRarity < 4 then return end

		-- Find the item ID to make sure it's not a ignored item
		local itemId = select(3, itemLink:find("item:(%d+):"))
		if not itemId then return end
		itemId = tonumber(itemId:trim())
		if type(itemId) ~= "number" or ignoredItems[itemId] then return end

		if type(itemName) == "string" and type(itemLink) == "string" and type(itemRarity) == "number" then
			self:ItemLooted(player, itemLink)
		end
	end
end

do
	local grouped = nil
	function BananaDKP:RAID_ROSTER_UPDATE()
		updateDisplayedNames = true
		local n = GetNumRaidMembers()
		if not grouped and n > 0 then
			grouped = true
			if db.mode == 1 then
				self:SendCommMessage("BDKP_SERVER_AVAILABLE", (UnitName("player")), "RAID")
			elseif db.mode == 2 and db.host then
				self:SendCommMessage("BDKP_CLIENT_AVAILABLE", (UnitName("player")), "WHISPER", db.host)
			end
		elseif grouped and n == 0 then
			self:StopCounter()
			self:ClearSitoutList()
			grouped = nil
		end
	end
	function BananaDKP:PARTY_MEMBERS_CHANGED()
		updateDisplayedNames = true
		local n = GetNumPartyMembers()
		if not grouped and n > 0 then
			grouped = true
			if db.mode == 1 then
				self:SendCommMessage("BDKP_SERVER_AVAILABLE", (UnitName("player")), "PARTY")
			elseif db.mode == 2 and db.host then
				self:SendCommMessage("BDKP_CLIENT_AVAILABLE", (UnitName("player")), "WHISPER", db.host)
			end
		elseif grouped and n == 0 and not UnitInRaid("player") then
			grouped = nil
			self:StopCounter()
			self:ClearSitoutList()
		end
	end
end

do
	local updating = nil
	function BananaDKP:GUILD_ROSTER_UPDATE()
		if updating then return end
		updating = true
		self:UnregisterEvent("GUILD_ROSTER_UPDATE")
		for k in pairs(guildRanks) do guildRanks[k] = nil end
		for i = 1, GuildControlGetNumRanks() do
			table.insert(guildRanks, GuildControlGetRankName(i))
		end
		local offline = GetGuildRosterShowOffline()
		local selection = GetGuildRosterSelection()
		SetGuildRosterShowOffline(true)
		local classes = {}
		for i = 1, GetNumGuildMembers() do
			local name, rank, _, _, localizedClass, _, _, _, _, _, class = GetGuildRosterInfo(i)
			if name then
				if rawget(coloredName, name) then
					coloredName[name] = nil
				end
				guildMemberList[name] = class
				guildMemberRanks[name] = rank
				guildMemberClasses[name] = localizedClass
				classes[localizedClass] = true
				if clients[name] then clients[name] = nil end
			end
		end
		SetGuildRosterShowOffline(offline)
		SetGuildRosterSelection(selection)

		for k in pairs(guildClasses) do guildClasses[k] = nil end
		for k in pairs(classes) do
			table.insert(guildClasses, k)
		end

		updateDisplayedNames = true
		updating = nil
		self:RegisterEvent("GUILD_ROSTER_UPDATE")
	end
end

-----------------------------------------------------------------------
-- Hooks
-----------------------------------------------------------------------

local function tablePrinter(t, recipient)
	for i, v in ipairs(t) do
		if type(v) == "string" then
			BananaDKP:Chat(L["BDKP: %s."]:format(v), "WHISPER", nil, recipient)
		elseif type(v) == "table" then
			tablePrinter(v, recipient)
		end
	end
end

local bidTimer = nil
local function bidTimeout()
	for k in pairs(tmpBidTable) do tmpBidTable[k] = nil end
	for k in pairs(bidders) do table.insert(tmpBidTable, k) end
	if #tmpBidTable > 0 then
		table.sort(tmpBidTable, dkpSorter)
		local winner = tmpBidTable[1]
		local winnerDkp = getDKP(winner)
		if #tmpBidTable > 1 and winnerDkp == getDKP(tmpBidTable[2]) then
			local more = true
			while more do
				table.remove(tmpBidTable, #tmpBidTable)
				more = getDKP(tmpBidTable[#tmpBidTable]) ~= winnerDkp
			end
			BananaDKP:Chat(L["Tie: %s."]:format(table.concat(tmpBidTable, ", ")), "RAID")
		else
			BananaDKP:Chat(L["%s wins (%d dkp)!"]:format(winner, winnerDkp), "RAID")
		end
	end
	takingBids = nil
	bidTimer = nil
end

local function newBidder(bidder)
	-- Force a DKP update in case we're in client mode.
	local dkp = getDKP(bidder)
	bidders[bidder] = true
	BananaDKP:CancelTimer(bidTimer, true)
	bidTimer = BananaDKP:ScheduleTimer(bidTimeout, 15)
end

chatHandlers.CHAT_MSG_WHISPER = function(msg, author)
	if db.mode == 2 and UnitInRaid(db.host) then return end
	if msg:find("^bdkp") then
		local command = select(3, msg:find("^bdkp%s?(.*)"))
		local input = command and command:len() > 0 and command:gsub("^%l", string.upper) or author
		local dkp = getDKP(input)
		if type(dkp) == "number" then
			BananaDKP:Chat(L["BDKP: DKP for %s: %s."]:format(input, tostring(dkp)), "WHISPER", nil, author)
		else
			BananaDKP:Chat(L["BDKP: Unknown command or player not found."], "WHISPER", nil, author)
		end
		return true
	elseif takingBids and msg:find("^bid") then
		newBidder(author)
		return true
	elseif msg == "sitout" then
		BananaDKP:ProcessSitoutWhisper(author)
		return true
	end
end

chatHandlers.CHAT_MSG_WHISPER_INFORM = function(msg, recipient)
	if msg:find("^BDKP") then return true end
end

chatHandlers.CHAT_MSG_RAID = function(msg, author)
	if db.mode == 2 then return end
	if author == playerName and msg:find(L["^bid:"]) then
		takingBids = true
		for k in pairs(bidders) do bidders[k] = nil end
	elseif takingBids and msg:find(L["^bid"]) then
		newBidder(author)
	end
end
chatHandlers.CHAT_MSG_RAID_LEADER = chatHandlers.CHAT_MSG_RAID

do
	local notSeenString = "^" .. _G.ERR_CHAT_PLAYER_NOT_FOUND_S:gsub("%%s", "(%%S+)") .. "$"
	local offlineString = "^" .. _G.ERR_FRIEND_OFFLINE_S:gsub("%%s", "(%%S+)") .. "$"
	chatHandlers.CHAT_MSG_SYSTEM = function(msg)
		local player = msg:match(offlineString)
		if not player then player = msg:match(notSeenString) end
		if player then
			if db.mode == 2 and db.host == player then
				if not hostOffline then
					print(L["Your host (%s) is offline. Please click the BananaDKP icon when he is online to continue updating."]:format(player))
					hostOffline = true
					BananaDKP:TriggerEvent("BananaDKP_HostOffline", true)
				end
				return true
			elseif db.mode == 1 and clients[player] then
				return true
			end
		end
	end
end

-----------------------------------------------------------------------
-- Communication Handling
-----------------------------------------------------------------------

function BananaDKP:Chat(text, chatType, language, destination)
	if ChatThrottleLib then
		ChatThrottleLib:SendChatMessage("BULK", "BananaDKP", text, chatType, language, destination)
	else
		SendChatMessage(text, chatType, language, destination)
	end
end

-----------------------------------------------------------------------
-- FuBar Handling
-----------------------------------------------------------------------

do
	local names = {}
	local offline = " %s (|cffff0000" .. PLAYER_OFFLINE .. "|r)"
	local offlineTip = L["|cffeda55fClick|r to attempt to contact the host again and continue updating."]
	local startTip = L["|cffeda55fClick|r to |cff00ff00start|r the hour counter."]
	local pauseTip = L["|cffeda55fClick|r to |cffff0000pause|r the hour counter."]
	
	local realm = GetRealmName()
	local function shouldShow(name)
		if guildMemberList[name] and
			db.disallowRanks[guildMemberRanks[name]] or
			db.disallowClasses[guildMemberClasses[name]] then
			return nil
		end
		-- Don't show players from other realms.
		local n, r = UnitName(name)
		-- Apparently the second return value can be "" for people from our own realm, not just nil, like wowwiki suggests.
		if n and r and r:trim():len() > 0 and r ~= realm then
			return nil
		end
		return true
	end
	
	function LDB.OnTooltipShow(tt)
		tt:AddLine("BananaDKP")
		if db.mode == 2 then
			if db.host then
				local h = coloredName[db.host]
				tt:AddDoubleLine(L["Host"], hostOffline and offline:format(h) or h, 1,1,0, 1,1,1)
			end
			tt:AddDoubleLine(L["Updated"], lastDKPUpdate and date("%H:%M", lastDKPUpdate) or L["Never"], 1,1,0, 1,1,1)
		end

		if updateDisplayedNames then
			for k in pairs(names) do names[k] = nil end
			if GetNumRaidMembers() > 0 then
				for i = 1, GetNumRaidMembers() do
					local n = GetRaidRosterInfo(i)
					if n and shouldShow(n) then
						table.insert(names, n)
					end
				end
			elseif GetNumPartyMembers() > 0 then
				table.insert(names, playerName)
				for i = 1, GetNumPartyMembers() do
					local n = UnitName("party" .. i)
					if n and shouldShow(n) then
						table.insert(names, n)
					end
				end
			else
				for k in pairs(guildMemberList) do
					if shouldShow(k) then
						local d = getDKP(k)
						if d then table.insert(names, k) end
					end
				end
			end
			updateDisplayedNames = nil

			table.sort(names, dkpSorter)
		end

		if #names > 0 then
			tt:AddLine(" ")
			tt:AddDoubleLine(L["Player"], L["DKP"], 1,1,1, 1,1,1)
			for i, v in ipairs(names) do
				local dkp = getDKP(v)
				if dkp and dkp > 0 then
					if v == playerName then
						tt:AddDoubleLine(v, dkp, 1,0,0, 1,0,0)
					else
						tt:AddDoubleLine(coloredName[v], dkp, 1,1,1, 1,1,1)
					end
				end
			end
		end

		if hostOffline then
			tt:AddLine(offlineTip, 0.2, 1, 0.2, 1)
		elseif db.mode == 1 and UnitInRaid("player") then
			if counterRunning then
				tt:AddLine(pauseTip, 0.2, 1, 0.2, 1)
			else
				tt:AddLine(startTip, 0.2, 1, 0.2, 1)
			end
		end
	end
end

-----------------------------------------------------------------------
-- Sitout list handling
-----------------------------------------------------------------------

function BananaDKP:ProcessSitoutWhisper(player)
	if not guildMemberList[player] then
		print(L["%q tried to add himself to our sitout list, but he's not in our guild."]:format(player))
		return
	end
	if db.sitouts[player] then
		self:RemoveSitoutPlayer(player)
	else
		self:AddSitoutPlayer(player)
	end
end

function BananaDKP:ClearSitoutList()
	for k in pairs(db.sitouts) do db.sitouts[k] = nil end
	self:RebuildSitoutList()
end

function BananaDKP:AddSitoutPlayer(name)
	db.sitouts[name] = true
	print(L["%q added to the sitout list."]:format(name))
	self:Chat(L["BDKP: You have been ADDED to the sitout list. Whisper me again to remove yourself."], "WHISPER", nil, name)
	self:RebuildSitoutList()
end

function BananaDKP:RemoveSitoutPlayer(name)
	db.sitouts[name] = nil
	print(L["%q removed from the sitout list."]:format(name))
	self:Chat(L["BDKP: You have been REMOVED from the sitout list. Whisper me again to add yourself."], "WHISPER", nil, name)
	self:RebuildSitoutList()
end

function BananaDKP:RebuildSitoutList()
	local opts = options.args.sitouts.args

	for k, v in pairs(opts) do
		if k ~= "add" and k ~= "clear" and k ~= "spacer" then
			opts[k].hidden = true
		end
	end

	for k, v in pairs(db.sitouts) do
		if not opts[k] then
			opts[k] = {
				type = "execute",
				name = coloredName[k],
				desc = L["Remove %s."]:format(k),
				func = "RemoveSitoutPlayer",
				passValue = k,
				order = 100,
			}
		end
		opts[k].hidden = nil
	end
end

